/* This is a generated file by rename_g2clib_symbols.h. *DO NOT EDIT MANUALLY !* */
#define cmplxpack gdal_cmplxpack
#define compack gdal_compack
#define comunpack gdal_comunpack
#define dec_png gdal_dec_png
#define DoubleToFloatClamp gdal_DoubleToFloatClamp
#define DoubleToFloatClamp gdal_DoubleToFloatClamp
#define DoubleToFloatClamp gdal_DoubleToFloatClamp
#define DoubleToFloatClamp gdal_DoubleToFloatClamp
#define DoubleToFloatClamp gdal_DoubleToFloatClamp
#define enc_jpeg2000 gdal_enc_jpeg2000
#define enc_png gdal_enc_png
#define extdrstemplate gdal_extdrstemplate
#define extgridtemplate gdal_extgridtemplate
#define extpdstemplate gdal_extpdstemplate
#define frame_dummy gdal_frame_dummy
#define g2_addfield gdal_g2_addfield
#define g2_addgrid gdal_g2_addgrid
#define g2_addlocal gdal_g2_addlocal
#define g2_create gdal_g2_create
#define g2_free gdal_g2_free
#define g2_getfld gdal_g2_getfld
#define g2_gribend gdal_g2_gribend
#define g2_info gdal_g2_info
#define g2_miss gdal_g2_miss
#define g2_unpack1 gdal_g2_unpack1
#define g2_unpack2 gdal_g2_unpack2
#define g2_unpack3 gdal_g2_unpack3
#define g2_unpack4 gdal_g2_unpack4
#define g2_unpack5 gdal_g2_unpack5
#define g2_unpack6 gdal_g2_unpack6
#define g2_unpack7 gdal_g2_unpack7
#define gbit gdal_gbit
#define gbit2 gdal_gbit2
#define gbits gdal_gbits
#define getdim gdal_getdim
#define getdrsindex gdal_getdrsindex
#define getdrstemplate gdal_getdrstemplate
#define getgridindex gdal_getgridindex
#define getgridtemplate gdal_getgridtemplate
#define getpdsindex gdal_getpdsindex
#define getpdstemplate gdal_getpdstemplate
#define getpoly gdal_getpoly
#define get_templatesdrs gdal_get_templatesdrs
#define get_templatesgrid gdal_get_templatesgrid
#define get_templatespds gdal_get_templatespds
#define int_power gdal_int_power
#define jpcpack gdal_jpcpack
#define jpcunpack gdal_jpcunpack
#define misspack gdal_misspack
#define mkieee gdal_mkieee
#define pack_gp gdal_pack_gp
#define pngpack gdal_pngpack
#define pngunpack gdal_pngunpack
#define rdieee gdal_rdieee
#define reduce gdal_reduce
#define sbit gdal_sbit
#define sbits gdal_sbits
#define seekgb gdal_seekgb
#define simpack gdal_simpack
#define simunpack gdal_simunpack
#define specpack gdal_specpack
#define specunpack gdal_specunpack
#define templatesdrs gdal_templatesdrs
#define templatesgrid gdal_templatesgrid
#define templatespds gdal_templatespds
