/* This is a generated file by dump_symbols.h. *DO NOT EDIT MANUALLY !* */
#define DefaultPrint gdal_DefaultPrint
#define DefaultRead gdal_DefaultRead
#define EPSGProjMethodToCTProjMethod gdal_EPSGProjMethodToCTProjMethod
#define FindCode gdal_FindCode
#define FindName gdal_FindName
#define FindTable gdal_FindTable
#define frame_dummy gdal_frame_dummy
#define GetNameFromDatabase gdal_GetNameFromDatabase
#define GTIFAllocDefn gdal_GTIFAllocDefn
#define GTIFAngleStringToDD gdal_GTIFAngleStringToDD
#define GTIFAngleToDD gdal_GTIFAngleToDD
#define _GTIFcalloc gdal__GTIFcalloc
#define GTIFDeaccessCSV gdal_GTIFDeaccessCSV
#define GTIFDecToDMS gdal_GTIFDecToDMS
#define GTIFDirectoryInfo gdal_GTIFDirectoryInfo
#define GTIFErrorFunction gdal_GTIFErrorFunction
#define GTIFFetchProjParms gdal_GTIFFetchProjParms
#define GTIFFree gdal_GTIFFree
#define _GTIFFree gdal__GTIFFree
#define GTIFFreeDefn gdal_GTIFFreeDefn
#define GTIFFreeMemory gdal_GTIFFreeMemory
#define GTIFGetDatumInfo gdal_GTIFGetDatumInfo
#define GTIFGetDatumInfoEx gdal_GTIFGetDatumInfoEx
#define GTIFGetDefn gdal_GTIFGetDefn
#define GTIFGetEllipsoidInfo gdal_GTIFGetEllipsoidInfo
#define GTIFGetEllipsoidInfoEx gdal_GTIFGetEllipsoidInfoEx
#define _GTIFGetField gdal__GTIFGetField
#define _GTIFGetField gdal__GTIFGetField
#define GTIFGetGCSInfo gdal_GTIFGetGCSInfo
#define GTIFGetGCSInfoEx gdal_GTIFGetGCSInfoEx
#define GTIFGetPCSInfo gdal_GTIFGetPCSInfo
#define GTIFGetPCSInfoEx gdal_GTIFGetPCSInfoEx
#define GTIFGetPMInfo gdal_GTIFGetPMInfo
#define GTIFGetPMInfoEx gdal_GTIFGetPMInfoEx
#define GTIFGetProj4Defn gdal_GTIFGetProj4Defn
#define GTIFGetProjTRFInfo gdal_GTIFGetProjTRFInfo
#define GTIFGetProjTRFInfoEx gdal_GTIFGetProjTRFInfoEx
#define GTIFGetUOMAngleInfo gdal_GTIFGetUOMAngleInfo
#define GTIFGetUOMAngleInfoEx gdal_GTIFGetUOMAngleInfoEx
#define GTIFGetUOMLengthInfo gdal_GTIFGetUOMLengthInfo
#define GTIFGetUOMLengthInfoEx gdal_GTIFGetUOMLengthInfoEx
#define GTIFGetUserData gdal_GTIFGetUserData
#define GTIFImageToPCS gdal_GTIFImageToPCS
#define GTIFImport gdal_GTIFImport
#define GTIFKeyCode gdal_GTIFKeyCode
#define GTIFKeyGet gdal_GTIFKeyGet
#define GTIFKeyGetASCII gdal_GTIFKeyGetASCII
#define GTIFKeyGetDOUBLE gdal_GTIFKeyGetDOUBLE
#define GTIFKeyGetInternal gdal_GTIFKeyGetInternal
#define GTIFKeyGetSHORT gdal_GTIFKeyGetSHORT
#define GTIFKeyGetSSHORT gdal_GTIFKeyGetSSHORT
#define GTIFKeyInfo gdal_GTIFKeyInfo
#define GTIFKeyName gdal_GTIFKeyName
#define GTIFKeyNameEx gdal_GTIFKeyNameEx
#define GTIFKeySet gdal_GTIFKeySet
#define GTIFMapSysToPCS gdal_GTIFMapSysToPCS
#define GTIFMapSysToProj gdal_GTIFMapSysToProj
#define _GTIFmemcpy gdal__GTIFmemcpy
#define GTIFNew gdal_GTIFNew
#define GTIFNewEx gdal_GTIFNewEx
#define GTIFNewSimpleTags gdal_GTIFNewSimpleTags
#define GTIFNewWithMethods gdal_GTIFNewWithMethods
#define GTIFNewWithMethodsEx gdal_GTIFNewWithMethodsEx
#define GTIFPCSToImage gdal_GTIFPCSToImage
#define GTIFPCSToMapSys gdal_GTIFPCSToMapSys
#define GTIFPrint gdal_GTIFPrint
#define GTIFPrintDefn gdal_GTIFPrintDefn
#define GTIFPrintDefnEx gdal_GTIFPrintDefnEx
#define GTIFProj4AppendEllipsoid gdal_GTIFProj4AppendEllipsoid
#define GTIFProj4FromLatLong gdal_GTIFProj4FromLatLong
#define GTIFProj4ToLatLong gdal_GTIFProj4ToLatLong
#define GTIFProjToMapSys gdal_GTIFProjToMapSys
#define _GTIFrealloc gdal__GTIFrealloc
#define _GTIFSetDefaultTIFF gdal__GTIFSetDefaultTIFF
#define _GTIFSetField gdal__GTIFSetField
#define _GTIFSetField gdal__GTIFSetField
#define GTIFSetFromProj4 gdal_GTIFSetFromProj4
#define GTIFSetSimpleTagsMethods gdal_GTIFSetSimpleTagsMethods
#define GTIFSetVersionNumbers gdal_GTIFSetVersionNumbers
#define GTIFTagCode gdal_GTIFTagCode
#define GTIFTagName gdal_GTIFTagName
#define _GTIFTagType gdal__GTIFTagType
#define _GTIFTagType gdal__GTIFTagType
#define GTIFTiepointTranslate gdal_GTIFTiepointTranslate
#define GTIFTypeCode gdal_GTIFTypeCode
#define GTIFTypeName gdal_GTIFTypeName
#define GTIFValueCode gdal_GTIFValueCode
#define GTIFValueName gdal_GTIFValueName
#define GTIFValueNameEx gdal_GTIFValueNameEx
#define GTIFWriteKeys gdal_GTIFWriteKeys
#define inv_geotransform gdal_inv_geotransform
#define OSRFreeStringList gdal_OSRFreeStringList
#define OSR_GDV gdal_OSR_GDV
#define OSR_GSV gdal_OSR_GSV
#define OSRProj4Tokenize gdal_OSRProj4Tokenize
#define PrintGeoTags gdal_PrintGeoTags
#define PrintKey gdal_PrintKey
#define PrintTag gdal_PrintTag
#define ReadKey gdal_ReadKey
#define ReadKey gdal_ReadKey
#define ReadTag gdal_ReadTag
#define SetCSVFilenameHook gdal_SetCSVFilenameHook
#define SetGTParamIds gdal_SetGTParamIds
#define SortKeys gdal_SortKeys
#define ST_Create gdal_ST_Create
#define ST_Destroy gdal_ST_Destroy
#define ST_GetKey gdal_ST_GetKey
#define StringError gdal_StringError
#define ST_SetKey gdal_ST_SetKey
#define ST_TagType gdal_ST_TagType
#define ST_TypeSize gdal_ST_TypeSize
#define WriteKey gdal_WriteKey
#define XTIFFClientOpen gdal_XTIFFClientOpen
#define XTIFFClose gdal_XTIFFClose
#define _XTIFFDefaultDirectory gdal__XTIFFDefaultDirectory
#define XTIFFFdOpen gdal_XTIFFFdOpen
#define XTIFFInitialize gdal_XTIFFInitialize
#define _XTIFFLocalDefaultDirectory gdal__XTIFFLocalDefaultDirectory
#define XTIFFOpen gdal_XTIFFOpen
#define StatePlaneTable gdal_StatePlaneTable
#define _keyInfo gdal__keyInfo
#define _keyInfoV11 gdal__keyInfoV11
#define _csdefaultValue gdal__csdefaultValue
#define _modeltypeValue gdal__modeltypeValue
#define _rastertypeValue gdal__rastertypeValue
#define _geounitsValue gdal__geounitsValue
#define _geographicValue gdal__geographicValue
#define _geodeticdatumValue gdal__geodeticdatumValue
#define _ellipsoidValue gdal__ellipsoidValue
#define _primemeridianValue gdal__primemeridianValue
#define _pcstypeValue gdal__pcstypeValue
#define _projectionValue gdal__projectionValue
#define _coordtransValue gdal__coordtransValue
#define _vertcstypeValue gdal__vertcstypeValue
#define _vdatumValue gdal__vdatumValue
#define _formatInfo gdal__formatInfo
#define _tagInfo gdal__tagInfo
#define xtiffFieldInfo gdal_xtiffFieldInfo
#define _gtiff_size gdal__gtiff_size
#define _ParentExtender gdal__ParentExtender
