#define compare_ints gdal_compare_ints
#define DBFAddField gdal_DBFAddField
#define DBFAddNativeFieldType gdal_DBFAddNativeFieldType
#define DBFAlterFieldDefn gdal_DBFAlterFieldDefn
#define DBFCloneEmpty gdal_DBFCloneEmpty
#define DBFClose gdal_DBFClose
#define DBFCreateEx gdal_DBFCreateEx
#define DBFCreate gdal_DBFCreate
#define DBFCreateLL gdal_DBFCreateLL
#define DBFDeleteField gdal_DBFDeleteField
#define DBFFlushRecord gdal_DBFFlushRecord
#define DBFGetCodePage gdal_DBFGetCodePage
#define DBFGetFieldCount gdal_DBFGetFieldCount
#define DBFGetFieldIndex gdal_DBFGetFieldIndex
#define DBFGetFieldInfo gdal_DBFGetFieldInfo
#define DBFGetNativeFieldType gdal_DBFGetNativeFieldType
#define DBFGetNullCharacter gdal_DBFGetNullCharacter
#define DBFGetRecordCount gdal_DBFGetRecordCount
#define DBFIsAttributeNULL gdal_DBFIsAttributeNULL
#define DBFIsRecordDeleted gdal_DBFIsRecordDeleted
#define DBFIsValueNULL gdal_DBFIsValueNULL
#define DBFLoadRecord gdal_DBFLoadRecord
#define DBFMarkRecordDeleted gdal_DBFMarkRecordDeleted
#define DBFOpen gdal_DBFOpen
#define DBFOpenLL gdal_DBFOpenLL
#define DBFReadAttribute gdal_DBFReadAttribute
#define DBFReadDoubleAttribute gdal_DBFReadDoubleAttribute
#define DBFReadIntegerAttribute gdal_DBFReadIntegerAttribute
#define DBFReadLogicalAttribute gdal_DBFReadLogicalAttribute
#define DBFReadStringAttribute gdal_DBFReadStringAttribute
#define DBFReadTuple gdal_DBFReadTuple
#define DBFReorderFields gdal_DBFReorderFields
#define DBFSetLastModifiedDate gdal_DBFSetLastModifiedDate
#define DBFSetWriteEndOfFileChar gdal_DBFSetWriteEndOfFileChar
#define DBFUpdateHeader gdal_DBFUpdateHeader
#define DBFWriteAttributeDirectly gdal_DBFWriteAttributeDirectly
#define DBFWriteAttribute gdal_DBFWriteAttribute
#define DBFWriteDoubleAttribute gdal_DBFWriteDoubleAttribute
#define DBFWriteHeader gdal_DBFWriteHeader
#define DBFWriteIntegerAttribute gdal_DBFWriteIntegerAttribute
#define DBFWriteLogicalAttribute gdal_DBFWriteLogicalAttribute
#define DBFWriteNULLAttribute gdal_DBFWriteNULLAttribute
#define DBFWriteStringAttribute gdal_DBFWriteStringAttribute
#define DBFWriteTuple gdal_DBFWriteTuple
#define frame_dummy gdal_frame_dummy
#define SBNAddShapeId gdal_SBNAddShapeId
#define SBNCloseDiskTree gdal_SBNCloseDiskTree
#define SBNOpenDiskTree gdal_SBNOpenDiskTree
#define SBNSearchDiskInternal gdal_SBNSearchDiskInternal
#define SBNSearchDiskTree gdal_SBNSearchDiskTree
#define SBNSearchDiskTreeInteger gdal_SBNSearchDiskTreeInteger
#define SBNSearchFreeIds gdal_SBNSearchFreeIds
#define SfRealloc gdal_SfRealloc
#define SHPAllocBuffer gdal_SHPAllocBuffer
#define SHPCheckBoundsOverlap gdal_SHPCheckBoundsOverlap
#define SHPCheckObjectContained gdal_SHPCheckObjectContained
#define SHPCloseDiskTree gdal_SHPCloseDiskTree
#define SHPClose gdal_SHPClose
#define SHPComputeExtents gdal_SHPComputeExtents
#define SHPCreate gdal_SHPCreate
#define SHPCreateLL gdal_SHPCreateLL
#define SHPCreateObject gdal_SHPCreateObject
#define SHPCreateSimpleObject gdal_SHPCreateSimpleObject
#define SHPCreateTree gdal_SHPCreateTree
#define SHPDestroyObject gdal_SHPDestroyObject
#define SHPDestroyTree gdal_SHPDestroyTree
#define SHPDestroyTreeNode gdal_SHPDestroyTreeNode
#define SHPGetInfo gdal_SHPGetInfo
#define SHPGetSubNodeOffset gdal_SHPGetSubNodeOffset
#define SHPOpenDiskTree gdal_SHPOpenDiskTree
#define SHPOpen gdal_SHPOpen
#define SHPOpenLLEx gdal_SHPOpenLLEx
#define SHPOpenLL gdal_SHPOpenLL
#define SHPPartTypeName gdal_SHPPartTypeName
#define SHPReadObject gdal_SHPReadObject
#define SHPReallocObjectBufIfNecessary gdal_SHPReallocObjectBufIfNecessary
#define SHPRestoreSHX gdal_SHPRestoreSHX
#define SHPRewindObject gdal_SHPRewindObject
#define SHPSearchDiskTreeEx gdal_SHPSearchDiskTreeEx
#define SHPSearchDiskTree gdal_SHPSearchDiskTree
#define SHPSearchDiskTreeNode gdal_SHPSearchDiskTreeNode
#define _SHPSetBounds gdal__SHPSetBounds
#define SHPSetFastModeReadObject gdal_SHPSetFastModeReadObject
#define SHPTreeAddShapeId gdal_SHPTreeAddShapeId
#define SHPTreeCollectShapeIds gdal_SHPTreeCollectShapeIds
#define SHPTreeFindLikelyShapes gdal_SHPTreeFindLikelyShapes
#define SHPTreeNodeAddShapeId gdal_SHPTreeNodeAddShapeId
#define SHPTreeNodeCreate gdal_SHPTreeNodeCreate
#define SHPTreeNodeTrim gdal_SHPTreeNodeTrim
#define SHPTreeReadLibc gdal_SHPTreeReadLibc
#define SHPTreeSeekLibc gdal_SHPTreeSeekLibc
#define SHPTreeSplitBounds gdal_SHPTreeSplitBounds
#define SHPTreeTrimExtraNodes gdal_SHPTreeTrimExtraNodes
#define SHPTypeName gdal_SHPTypeName
#define SHPWriteHeader gdal_SHPWriteHeader
#define SHPWriteObject gdal_SHPWriteObject
#define SHPWriteTree gdal_SHPWriteTree
#define SHPWriteTreeLL gdal_SHPWriteTreeLL
#define SHPWriteTreeNode gdal_SHPWriteTreeNode
#define str_to_upper gdal_str_to_upper
#define SwapWord gdal_SwapWord
/* This is a generated file by dump_symbols.h. *DO NOT EDIT MANUALLY !* */
