/**
  * Provides the classes necessary to handle OGR vector datasources.
  *
  * This package provides the classes for the Driver, DataSource, Layer, Feature, Geometry... objects that are used for handling OGR vector datasources.
  */
package org.gdal.ogr;